;;; imap.el --- imap library
;; Copyright (C) 1998 Free Software Foundation, Inc.

;; Author: Simon Josefsson <jas@pdc.kth.se>
;; Keywords: mail

;; This file is not part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;;
;; Without the work of John McClary Prevost and Jim Radford this library
;; would not have seen the light of day. Many thanks.
;;
;; This is a elisp library providing an interface for talking to IMAP
;; server. 
;;
;; It supports RFC1730/2060 (IMAP4/IMAP4rev1), implemented imap
;; extensions are RFC2195 (CRAM-MD5), RFC2086 (ACL), RFC2342
;; (NAMESPACE), RFC2359 (UIDPLUS), the kerberos V4 part of RFC1731.
;; It also support the UNSELECT extension of the Cyrus IMAPD.
;;
;; Example session:
;;
;; (imap-open "my.mail.server")
;; " *imap* my.mail.server:0"
;;
;; (imap-authenticate " *imap* my.mail.server:0")
;; auth
;;
;; (imap-mailbox-lsub " *imap* my.mail.server:0")
;; ("INBOX.sentmail" "INBOX.private" "INBOX.draft" "INBOX.spam")
;;
;; (imap-mailbox-list " *imap* my.mail.server:0" "INBOX.n%")
;; ("INBOX.ntbugtraq" "INBOX.nymb" "INBOX.netboot" "INBOX.nnimap")
;;
;; (imap-mailbox-select "INBOX.nnimap" " *imap* my.mail.server:0")
;; "INBOX.nnimap"
;;
;; (imap-mailbox-get 'exists nil " *imap* my.mail.server:0")
;; 39
;;
;; (imap-search "FLAGGED SINCE 18-DEC-98" " *imap* my.mail.server:0")
;; (235 236)
;;
;; xxx: more examples
;;
;; Server functions:
;;
;; (imap-open SERVER &optional PORT STREAM AUTH BUFFER)
;;    
;;    Open a IMAP connection to host SERVER at PORT returning a
;;    buffer. If PORT is unspecified, a default value is used (143 except
;;    for SSL which use 993).
;;    STREAM indicates the stream to use, see `imap-streams' for available
;;    streams. If nil, it choices the best stream the server is capable of.
;;    AUTH indicates authenticator to use, see `imap-authenticators' for
;;    available authenticators. If nil, it choices the best stream the
;;    server is capable of.
;;    BUFFER can be a buffer or a name of a buffer, which is created if
;;    necessery.
;;
;; (imap-opened &optional BUFFER)
;;
;;    Return non-nil if connection to imap server in BUFFER is open. If
;;    BUFFER is nil then the current buffer is used.
;;
;; (imap-close &optional BUFFER)
;;
;;    Close connection to server in BUFFER. If BUFFER is nil, the current
;;    buffer is used.
;;
;; (imap-authenticate BUFFER &optional USER PASSWD)
;;
;;    Authenticate to server in BUFFER, using current buffer if nil. It
;;    uses the authenticator specified when opening the server. If the
;;    authenticator requires username/passwords, they are queried from the
;;    user and optionally stored in the buffer.  If USER and/or PASSWD is
;;    specified, the user will not be questioned and the username and/or
;;    password is remembered in the buffer.
;;
;; (imap-capability &optional IDENTIFIER BUFFER)
;;
;;    Return a list of identifiers which server in BUFFER support. If
;;    IDENTIFIER, return non-nil if it's among the servers capabilities. If
;;    BUFFER is nil, the current buffer is assumed.
;;
;; (imap-namespace &optional BUFFER)
;;
;;    Return a namespace hierarchy at server in BUFFER. If BUFFER is nil,
;;    the current buffer is assumed.
;;
;; imap-mailbox-put
;; imap-mailbox-get
;; imap-mailbox-map
;; imap-mailbox-select
;; imap-mailbox-unselect
;; imap-mailbox-close
;; imap-mailbox-lsub
;; imap-mailbox-list
;; imap-mailbox-subscribe
;; imap-mailbox-unsubscribe
;; imap-message-put
;; imap-message-get
;; imap-message-map
;; imap-search
;; imap-message-flags-set
;; imap-message-flags-del
;; imap-message-flags-add
;; imap-wait-for-tag
;; imap-ok-p
;; imap-send-command-wait
;; 
;; Todo:
;; 
;; o Handle literals in ENVELOPE/BODY fetch responses
;; o Parse UIDs as strings? (28 bit limit)
;; o Sleep.
;;

;;; Code:

(eval-when-compile (require 'cl))

(eval-and-compile
  (autoload 'open-ssl-stream "ssl")
  (autoload 'base64-decode "base64")
  (autoload 'rfc2104-hash "rfc2104")
  (autoload 'md5 "md5")
  (unless (fboundp 'open-network-stream)
    (require 'tcp)))

;; User variables.

(defvar imap-imtest-arguments "-kp"
  "Privacy / integrity checking arguments passed to `imtest'.")

(defvar imap-default-user (user-login-name)
  "Default username to use.")

(defvar imap-error nil
  "Error codes from the last command.")

;; Various variables.

(defvar imap-fetch-data-hook nil
  "Hooks called after receiving each FETCH response.")

(defvar imap-streams '(kerberos4 ssl network)
  "Priority of streams to consider when opening connection to
server.")

(defvar imap-stream-alist
  '((kerberos4 imap-kerberos4s-p imap-kerberos4-open)
    (ssl       imap-ssl-p        imap-ssl-open)
    (network   imap-network-p    imap-network-open))
  "Definition of network streams.

(NAME CHECK OPEN)

NAME names the stream, CHECK is a function returning non-nil if the
server support the stream and OPEN is a function for opening the
stream.")

(defvar imap-authenticators '(kerberos4 cram-md5 login anonymous)
  "Priority of authenticators to consider when authenticating to
server.")

(defvar imap-authenticator-alist 
  '((kerberos4 imap-kerberos4a-p imap-kerberos4-auth)
    (cram-md5  imap-cram-md5-p   imap-cram-md5-auth)
    (login     imap-login-p      imap-login-auth)
    (anonymous imap-anonymous-p  imap-anonymous-auth))
  "Definition of authenticators.

(NAME CHECK AUTHENTICATE)

NAME names the authenticator. CHECK is a function returning non-nil if
the server support the authenticator and AUTHENTICATE is a function
for doing the actuall authentification.")

;; Internal constants. Change theese and die.

(defconst imap-default-port 143)
(defconst imap-default-ssl-port 993)
(defconst imap-default-stream 'network)
(defconst imap-coding-system-for-read 'binary)
(defconst imap-coding-system-for-write 'binary)
(defconst imap-local-variables '(imap-server
				 imap-port
				 imap-client-eol
				 imap-server-eol
				 imap-auth
				 imap-stream
				 imap-username
				 imap-password
				 imap-current-mailbox
				 imap-current-target-mailbox
				 imap-capability
				 imap-namespace
				 imap-state
				 imap-reached-tag
				 imap-failed-tags
				 imap-tag
				 imap-process
				 imap-mailbox-data))

;; Internal variables.

(defvar imap-stream nil)
(defvar imap-auth nil)
(defvar imap-server nil)
(defvar imap-port nil)
(defvar imap-username nil)
(defvar imap-password nil)
(defvar imap-state 'closed 
  "IMAP state. Valid states are `closed', `initial', `nonauth',
`auth', `selected' and `examine'.")

(defvar imap-server-eol "\r\n"
  "The EOL string sent from the server.")

(defvar imap-client-eol "\r\n"
  "The EOL string we send to the server.")

(defvar imap-current-mailbox nil
  "Current mailbox name.")

(defvar imap-current-target-mailbox nil
  "Current target mailbox for COPY and APPEND commands.")

(defvar imap-mailbox-data nil
  "Obarray with mailbox data.")

(defvar imap-mailbox-prime 127
  "Length of imap-mailbox-data.")

(defvar imap-current-message nil
  "Current message number.")

(defvar imap-message-data nil
  "Obarray with message data.")

(defvar imap-message-prime 67
  "Length of imap-message-data.")

(defvar imap-capability nil
  "Capability for server.")

(defvar imap-namespace nil
  "Namespace for current server.")

(defvar imap-reached-tag 0
  "Lower limit on command tags that have been parsed.")

(defvar imap-failed-tags nil 
  "Alist of tags that failed. Each element is a list with four
elements; tag (a integer), response state (a symbol, `OK', `NO' or
`BAD'), response code (a string), and human readable response text (a
string).")

(defvar imap-tag 0
  "Command tag number.")

(defvar imap-process nil
  "Process.")

(defvar imap-log "*imap-log*"
  "Imap session trace.")

(defvar imap-debug nil;"*imap-debug*"
  "Random debug spew.")


;; Utility functions:

(defsubst imap-disable-multibyte ()
  "Enable multibyte in the current buffer."
  (when (fboundp 'set-buffer-multibyte)
    (set-buffer-multibyte nil)))

(defun imap-encode-string (string)
  ;; xxx make literal if strange characters in string
  (concat "\"" string "\""))

(defun imap-read-passwd (prompt &rest args)
  "Read a password using PROMPT. If ARGS, PROMPT is used as an
argument to `format'."
  (let ((prompt (if args
		    (apply 'format prompt args)
		  prompt)))
    (funcall (if (load "passwd" t)
		 'read-passwd
	       (unless (fboundp 'ange-ftp-read-passwd)
		 (autoload 'ange-ftp-read-passwd "ange-ftp"))
	       'ange-ftp-read-passwd) prompt)))

(defsubst imap-ok-p (status)
  (if (eq status 'OK)
      t
    (setq imap-error status)
    nil))


;; Server functions; stream stuff:

(defun imap-kerberos4s-p (buffer)
  (imap-capability 'AUTH=KERBEROS_V4 buffer))

(defun imap-kerberos4-open (name buffer server port)
  (message "Opening Kerberized IMAP connection...")
  (let* ((port (or port imap-default-port))
	 (coding-system-for-read imap-coding-system-for-read)
	 (coding-system-for-write imap-coding-system-for-write)
	 (process (start-process name buffer "imtest" imap-imtest-arguments
				 server (number-to-string port))))
    (when process
      (with-current-buffer buffer
	(setq imap-client-eol "\n")
	;; Result of authentication is a string: __Full privacy protection__
	(while (and (memq (process-status process) '(open run))
		    (goto-char (point-min))
		    (not (and (imap-parse-greeting)
			      (re-search-forward "__\\(.*\\)__\n" nil t))))
	  (accept-process-output process 1)
	  (sit-for 1))
	(and imap-log
	     (with-current-buffer (get-buffer-create imap-log)
	       (imap-disable-multibyte)
	       (buffer-disable-undo)
	       (goto-char (point-max))
	       (insert-buffer buffer)))
      (let ((response (match-string 1)))
	(erase-buffer)
	(message "Kerberized IMAP connection: %s" response)
	(if (and response (let ((case-fold-search nil))
			    (not (string-match "failed" response))))
	    process
	  (if (memq (process-status process) '(open run))
	      (imap-send-command-wait "LOGOUT"))
	  (delete-process process)
	  nil))))))
  
(defun imap-ssl-p (buffer)
  nil)

(defun imap-ssl-open-1 (name buffer server port extra-arg)
  (let* ((port (or port imap-default-ssl-port))
	 (coding-system-for-read imap-coding-system-for-read)
	 (coding-system-for-write imap-coding-system-for-write)
	 (ssl-program-arguments (list extra-arg "-connect" 
				      (format "%s:%d" server port)))
	 (process (open-ssl-stream name buffer server port)))
    (when process
      (with-current-buffer buffer
	(goto-char (point-min))
	(while (and (memq (process-status process) '(open run))
		    (goto-char (point-max))
		    (forward-line -1)
		    (not (imap-parse-greeting)))
	  (accept-process-output process 1)
	  (sit-for 1))
	(and imap-log
	     (with-current-buffer (get-buffer-create imap-log)
	       (imap-disable-multibyte)
	       (buffer-disable-undo)
	       (goto-char (point-max))
	       (insert-buffer buffer)))
	(erase-buffer))
      (when (memq (process-status process) '(open run))
	process))))

(defun imap-ssl-open (name buffer server port)
  (message "Opening SSL3 IMAP connection...")
  (let ((ret (imap-ssl-open-1 name buffer server port "-ssl3")))
    (if ret
	ret
      (message "Opening SSL2 IMAP connection...")
      (imap-ssl-open-1 name buffer server port "-ssl2"))))

(defun imap-network-p (buffer)
  t)

(defun imap-network-open (name buffer server port)
  (let* ((port (or port imap-default-port))
	 (coding-system-for-read imap-coding-system-for-read)
	 (coding-system-for-write imap-coding-system-for-write)
	 (process (open-network-stream name buffer server port)))
    (when process
      (while (and (memq (process-status process) '(open run))
		  (goto-char (point-min))
		  (not (imap-parse-greeting)))
	(accept-process-output process 1)
	(sit-for 1))
      (and imap-log
	   (with-current-buffer (get-buffer-create imap-log)
	     (imap-disable-multibyte)
	     (buffer-disable-undo)
	     (goto-char (point-max))
	     (insert-buffer buffer)))
      (when (memq (process-status process) '(open run))
	process))))
  
;; Server functions; authenticator stuff:

(defun imap-interactive-login (buffer loginfunc)
  "Login to server in BUFFER. LOGINFUNC is passed a username and a
password, it should return t if it where sucessful authenticating
itself to the server, nil otherwise. Returns t if login was
successful, nil otherwise."
  (with-current-buffer buffer
    (make-variable-buffer-local 'imap-username)
    (make-variable-buffer-local 'imap-password)
    (let (user passwd ret)
;;      (condition-case ()
	  (while (or (not user) (not passwd))
	    (setq user (or imap-username
			   (read-from-minibuffer 
			    (concat "IMAP username for " imap-server ": ")
			    (or user imap-default-user))))
	    (setq passwd (or imap-password
			     (imap-read-passwd
			      (concat "IMAP password for " user "@" 
				      imap-server ": "))))
	    (when (and user passwd)
	      (if (funcall loginfunc user passwd)
		  (progn
		    (setq ret t
			  imap-username user)
		    (if (and (not imap-password)
			     (y-or-n-p "Store password for this session? "))
			(setq imap-password passwd)))
		(message "Login failed...")
		(setq passwd nil)
		(sit-for 1))))
;;	(quit (with-current-buffer buffer
;;		(setq user nil
;;		      passwd nil)))
;;	(error (with-current-buffer buffer
;;		 (setq user nil
;;		       passwd nil))))
      ret)))

(defun imap-kerberos4a-p (buffer)
  (imap-capability 'AUTH=KERBEROS_V4 buffer))

(defun imap-kerberos4-auth (buffer)
  (eq imap-stream 'kerberos4))

(defun imap-cram-md5-p (buffer)
  (imap-capability 'AUTH=CRAM-MD5 buffer))

(defun imap-cram-md5-auth (buffer)
  "Login to server using the AUTH CRAM-MD5 method."
  (imap-interactive-login
   buffer
   (lambda (user passwd)
     (imap-ok-p
      (imap-send-command-wait
       (list
	"AUTHENTICATE CRAM-MD5"
	(lambda (challenge)
	  (let* ((decoded (base64-decode challenge))
		 (hash (rfc2104-hash 'md5 64 16 passwd decoded))
		 (response (concat user " " hash))
		 (encoded (base64-encode response)))
	    encoded))))))))

(defun imap-login-p (buffer)
  (not (imap-capability 'X-LOGIN-CMD-DISABLED buffer)))

(defun imap-login-auth (buffer)
  "Login to server using the LOGIN command."
  (imap-interactive-login buffer 
			  (lambda (user passwd)
			    (imap-ok-p (imap-send-command-wait 
					(concat "LOGIN \"" user "\" \"" 
						passwd "\""))))))

(defun imap-anonymous-p (buffer)
  t)

(defun imap-anonymous-open (buffer)
  (with-current-buffer buffer
    (imap-ok-p (imap-send-command-wait
		(concat "LOGIN anonymous \"" (concat (user-login-name) "@" 
						     (system-name)) "\"")))))

;; Server functions:

(defun imap-open-1 (buffer)
  (with-current-buffer buffer
    (erase-buffer)
    (setq imap-current-mailbox nil
	  imap-current-message nil
	  imap-state 'initial
	  imap-process (condition-case ()
			   (funcall (nth 2 (assq imap-stream 
						 imap-stream-alist))
				    "imap" buffer imap-server imap-port)
			 ((error quit) nil)))
    (when imap-process
      (set-process-filter imap-process 'imap-arrival-filter)
      (set-process-sentinel imap-process 'imap-sentinel)
      (while (and (eq imap-state 'initial)
		  (memq (process-status imap-process) '(open run)))
	(message "Waiting for response from %s..." imap-server)
	(accept-process-output imap-process 1))
      (message "Waiting for response from %s...done" imap-server)
      (and (memq (process-status imap-process) '(open run))
	   imap-process))))

(defun imap-open (server &optional port stream auth buffer)
  "Open a IMAP connection to host SERVER at PORT returning a
buffer. If PORT is unspecified, a default value is used (143 except
for SSL which use 993).
STREAM indicates the stream to use, see `imap-streams' for available
streams. If nil, it choices the best stream the server is capable of.
AUTH indicates authenticator to use, see `imap-authenticators' for
available authenticators. If nil, it choices the best stream the
server is capable of.
BUFFER can be a buffer or a name of a buffer, which is created if
necessery. If nil, the buffer name is generated."
  (setq buffer (or buffer (format " *imap* %s:%d" server (or port 0))))
  (with-current-buffer (get-buffer-create buffer)
    (if (imap-opened buffer)
	(imap-close buffer))
    (mapc 'make-variable-buffer-local imap-local-variables)
    (imap-disable-multibyte)
    (buffer-disable-undo)
    (setq imap-server (or server imap-server))
    (setq imap-port (or port imap-port))
    (setq imap-auth (or auth imap-auth))
    (setq imap-stream (or stream imap-stream))
    (when (let ((imap-stream (or imap-stream imap-default-stream)))
	    (imap-open-1 buffer))
      ;; Choose stream.
      (let (stream-changed)
	(when (null imap-stream)
	  (let ((streams imap-streams))
	    (while (setq stream (pop streams))
	      (if (funcall (nth 1 (assq stream imap-stream-alist)) buffer)
		  (setq stream-changed (not (eq (or imap-stream 
						    imap-default-stream)
						stream))
			imap-stream stream
			streams nil)))
	    (unless imap-stream
	      (error "Couldn't figure out a stream for server"))))
	(when stream-changed
	  (message "Reconnecting with %s..." imap-stream)
	  (imap-close buffer)
	  (imap-open-1 buffer)
	  (setq imap-capability nil)))
      (if (imap-opened buffer)
	  ;; Choose authenticator
	  (when (null imap-auth)
	    (let ((auths imap-authenticators))
	      (while (setq auth (pop auths))
		(if (funcall (nth 1 (assq auth imap-authenticator-alist)) 
			     buffer)
		    (setq imap-auth auth
			  auths nil)))
	      (unless imap-auth
		(error "Couldn't figure out authenticator for server")))))))
  (if (imap-opened buffer)
      buffer))

(defun imap-opened (&optional buffer)
  "Return non-nil if connection to imap server in BUFFER is open. If
BUFFER is nil then the current buffer is used."
  (and (setq buffer (get-buffer (or buffer (current-buffer))))
       (buffer-live-p buffer)
       (with-current-buffer buffer
	 (and imap-process
	      (memq (process-status imap-process) '(open run))))))

(defun imap-authenticate (&optional buffer user passwd)
  "Authenticate to server in BUFFER, using current buffer if nil. It
uses the authenticator specified when opening the server. If the
authenticator requires username/passwords, they are queried from the
user and optionally stored in the buffer.  If USER and/or PASSWD is
specified, the user will not be questioned and the username and/or
password is remembered in the buffer."
  (with-current-buffer (or buffer (current-buffer))
    (when (eq imap-state 'nonauth)
      (make-variable-buffer-local 'imap-username)
      (make-variable-buffer-local 'imap-password)
      (if user (setq imap-username user))
      (if passwd (setq imap-password passwd))
      (if (funcall (nth 2 (assq imap-auth imap-authenticator-alist)) buffer)
	  (setq imap-state 'auth)))))

(defun imap-close (&optional buffer)
  "Close connection to server in BUFFER. If BUFFER is nil, the current
buffer is used."
  (with-current-buffer (or buffer (current-buffer))
    (and (imap-opened)
	 (not (imap-ok-p (imap-send-command-wait "LOGOUT")))
	 (message "Server %s didn't let me log out" imap-server))
    (when (and imap-process
	       (memq (process-status imap-process) '(open run)))
      (delete-process imap-process))
    (setq imap-current-mailbox nil
	  imap-current-message nil
	  imap-process nil)
    (erase-buffer)
    t))

(defun imap-capability (&optional identifier buffer)
  "Return a list of identifiers which server in BUFFER support. If
IDENTIFIER, return non-nil if it's among the servers capabilities. If
BUFFER is nil, the current buffer is assumed."
  (with-current-buffer (or buffer (current-buffer))
    (unless imap-capability
      (unless (imap-ok-p (imap-send-command-wait "CAPABILITY"))
	(setq imap-capability '(IMAP2))))
    (if identifier
	(memq identifier imap-capability)
      imap-capability)))

(defun imap-namespace (&optional buffer)
  "Return a namespace hierarchy at server in BUFFER. If BUFFER is nil,
the current buffer is assumed."
  (with-current-buffer (or buffer (current-buffer))
    (unless imap-namespace
      (when (imap-capability 'NAMESPACE)
	(imap-send-command-wait "NAMESPACE")))
    imap-namespace))

(defun imap-send-command-wait (command &optional buffer)
  (imap-wait-for-tag (imap-send-command command buffer) buffer))


;; Mailbox functions:

(defun imap-mailbox-put (propname value &optional mailbox buffer)
  (with-current-buffer (or buffer (current-buffer))
    (put (intern (or mailbox imap-current-mailbox) imap-mailbox-data)
	 propname value)
    t))

(defun imap-mailbox-get (propname &optional mailbox buffer)
  (with-current-buffer (or buffer (current-buffer))
    (get (intern-soft (or mailbox imap-current-mailbox) imap-mailbox-data)
	 propname)))

(defun imap-mailbox-map (func &optional buffer)
  "Map a function across each mailbox in `imap-mailbox-data',
returning a list."
  (with-current-buffer (or buffer (current-buffer))
    (let (result)
      (mapatoms 
       (lambda (s)
	 (push (funcall func (symbol-name s)) result))
       imap-mailbox-data)
      result)))

(defun imap-mailbox-select (mailbox &optional buffer examine)
  (with-current-buffer (or buffer (current-buffer))
    (unless (and (string= mailbox imap-current-mailbox)
		 (or (and examine
			  (eq imap-state 'examine))
		     (and (not examine)
			  (eq imap-state 'selected))))
      (setq imap-current-mailbox mailbox)
      (if (imap-ok-p (imap-send-command-wait
		      (concat (if examine "EXAMINE" "SELECT") " " mailbox)))
	  (setq imap-message-data (make-vector imap-message-prime 0)
		imap-state (if examine 'examine 'selected))
	;; Failed SELECT unselects the current group
	(setq imap-current-mailbox nil
	      imap-message-data nil
	      imap-state 'auth)))
    imap-current-mailbox))

(defun imap-mailbox-unselect (&optional buffer)
  "Close current folder in BUFFER, without expunging articles."
  (with-current-buffer (or buffer (current-buffer))
    (when (or (eq imap-state 'auth)
	      (and (imap-capability 'UNSELECT)
		   (imap-ok-p (imap-send-command-wait "UNSELECT")))
	      (and (imap-ok-p 
		    (imap-send-command-wait (concat "EXAMINE "
						    imap-current-mailbox)))
		   (imap-ok-p (imap-send-command-wait "CLOSE"))))
      (setq imap-current-mailbox nil
	    imap-message-data nil
	    imap-state 'auth)
      t)))

(defun imap-mailbox-close (&optional buffer)
  "Expunge articles and close current folder in BUFFER."
  (with-current-buffer (or buffer (current-buffer))
    (when (imap-ok-p (imap-send-command-wait "CLOSE"))
      (setq imap-current-mailbox nil
	    imap-message-data nil
	    imap-state 'auth)
      t)))

(defun imap-mailbox-create (mailbox &optional buffer)
  "Create MAILBOX on server in BUFFER, if nil current buffer is assumed."
  (with-current-buffer (or buffer (current-buffer))
    (imap-ok-p 
     (imap-send-command-wait (list "CREATE " mailbox)))))

(defun imap-mailbox-lsub (&optional buffer reference)
  "Clear the mailbox data and fill it with subscribed mailboxes on
server in BUFFER. REFERENCE is the implementation-specific string that
has to be passed to LSUB."
  (with-current-buffer (or buffer (current-buffer))
    (setq imap-mailbox-data (make-vector imap-mailbox-prime 0))
    (when (imap-ok-p (imap-send-command-wait 
		      (concat "LSUB \"" reference "\" \"*\"")))
      (imap-mailbox-map 'identity))))

(defun imap-mailbox-list (&optional buffer root have-delimiter reference)
  "List all mailboxes that starts with ROOT in BUFFER. If
HAVE-DELIMITER is non-nil, a hierarchy delimiter is not added to
ROOT. REFERENCE is the implementation-specific string that has to be
passed to LIST."
  (with-current-buffer (or buffer (current-buffer))
    (setq imap-mailbox-data (make-vector imap-mailbox-prime 0))
    ;; Find hierarchy separator if root turns out to be a mailbox
    ;; instead of just a prefix.
    (when root
      (unless have-delimiter
	(imap-send-command-wait (concat "LIST \"" reference "\" \"" root "\""))))
    (when (imap-ok-p 
	   (imap-send-command-wait 
	    (concat "LIST \"" reference "\" \"" root
		    (when (and root (not have-delimiter))
		      (imap-mailbox-get 'delimiter root))
		    "%\"")))
      (imap-mailbox-map 'identity))))

(defun imap-mailbox-subscribe (mailbox &optional buffer)
  "Send the SUBSCRIBE command on the mailbox to server in
BUFFER. Returns non-nil if successful."
  (with-current-buffer (or buffer (current-buffer))
    (imap-ok-p (imap-send-command-wait (concat "SUBSCRIBE " mailbox)))))

(defun imap-mailbox-unsubscribe (mailbox &optional buffer)
  "Send the SUBSCRIBE command on the mailbox to server in
BUFFER. Returns non-nil if successful."
  (with-current-buffer (or buffer (current-buffer))
    (imap-ok-p (imap-send-command-wait (concat "UNSUBSCRIBE " mailbox)))))

(defun imap-mailbox-status (mailbox items &optional buffer)
  "Get status items ITEM in MAILBOX from server in BUFFER. ITEMS can
be a symbol or a list of symbols, valid symbols are one of the STATUS
data items -- ie 'messages, 'recent, 'uidnext, 'uidvalidity or
'unseen. If ITEMS is a list of symbols, a list of values is returned,
if ITEMS is a symbol only it's value is returned."
  (with-current-buffer (or buffer (current-buffer))
    (when (imap-ok-p 
	   (imap-send-command-wait (list "STATUS "
					 (imap-encode-string mailbox)
					 " "
					 (format "%s"
						 (if (listp items)
						     items 
						   (list items))))))
      (if (listp items)
	  (mapcar (lambda (item)
		    (imap-mailbox-get item mailbox))
		  items)
	(imap-mailbox-get items mailbox)))))

(defun imap-mailbox-acl-get (&optional mailbox buffer)
  "Get ACL on mailbox from server in BUFFER."
  (with-current-buffer (or buffer (current-buffer))
    (when (imap-ok-p
	   (imap-send-command-wait (list "GETACL "
					 (or mailbox imap-current-mailbox))))
      (imap-mailbox-get 'acl (or mailbox imap-current-mailbox)))))

(defun imap-mailbox-acl-set (identifier rights &optional mailbox buffer)
  "Change/set ACL for IDENTIFIER to RIGHTS in MAILBOX from server in
BUFFER."
  (with-current-buffer (or buffer (current-buffer))
    (imap-ok-p
     (imap-send-command-wait (list "SETACL "
				   (or mailbox imap-current-mailbox)
				   " "
				   identifier
				   " "
				   rights)))))

(defun imap-mailbox-acl-delete (identifier &optional mailbox buffer)
  "Removes any <identifier,rights> pair for IDENTIFIER in MAILBOX from
server in BUFFER."
  (with-current-buffer (or buffer (current-buffer))
    (imap-ok-p
     (imap-send-command-wait (list "DELETEACL "
				   (or mailbox imap-current-mailbox)
				   " "
				   identifier)))))


;; Message functions:

(defun imap-message-put (uid propname value &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (put (intern (number-to-string uid) imap-message-data)
	 propname value)
    t))

(defun imap-message-get (uid propname &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (get (intern-soft (number-to-string uid) imap-message-data)
	 propname)))

(defun imap-message-map (func propname &optional buffer)
  "Map a function across each mailbox in `imap-message-data',
returning a list."
  (with-current-buffer (or buffer (current-buffer))
    (let (result)
      (mapatoms
       (lambda (s)
	 (push (funcall func (get s 'UID) (get s propname)) result))
       imap-message-data)
      result)))

(defun imap-search (predicate &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (imap-mailbox-put 'search nil)
    (when (imap-ok-p (imap-send-command-wait (concat "UID SEARCH " predicate)))
      (imap-mailbox-get 'search))))

(defun imap-message-flags-set (articles flags &optional buffer silent)
  (when (and articles flags)
    (with-current-buffer (or buffer (current-buffer))
      (imap-ok-p (imap-send-command-wait
		  (concat "UID STORE " articles
			  " FLAGS" (if silent ".SILENT") " (" flags ")"))))))

(defun imap-message-flags-del (articles flags &optional buffer silent)
  (when (and articles flags)
    (with-current-buffer (or buffer (current-buffer))
      (imap-ok-p (imap-send-command-wait
		  (concat "UID STORE " articles
			  " -FLAGS" (if silent ".SILENT") " (" flags ")"))))))

(defun imap-message-flags-add (articles flags &optional buffer silent)
  (when (and articles flags)
    (with-current-buffer (or buffer (current-buffer))
      (imap-ok-p (imap-send-command-wait
		  (concat "UID STORE " articles
			  " +FLAGS" (if silent ".SILENT") " (" flags ")"))))))

(defun imap-message-copy (articles mailbox &optional buffer dont-create)
  "Copy ARTICLES to MAILBOX on server in BUFFER, creating mailbox if
it doesn't exist. If dont-create is non-nil, it will not create a
mailbox."
  (when articles
    (with-current-buffer (or buffer (current-buffer))
      (let ((cmd (list "UID COPY " articles " " mailbox))
	    (imap-current-target-mailbox mailbox))
	(if (imap-ok-p (imap-send-command-wait cmd))
	    t
	  (when (and (not dont-create) (imap-mailbox-get 'trycreate mailbox))
	    (imap-mailbox-create mailbox)
	    (imap-ok-p (imap-send-command-wait cmd))))))))

(defun imap-message-append (mailbox article &optional buffer flags date-time)
  "Append ARTICLE buffer to MAILBOX on server in BUFFER. FLAGS and
DATE-TIME is currently not used."
  (with-current-buffer (or buffer (current-buffer))
    (if (imap-capability 'UIDPLUS)
	(let ((imap-current-target-mailbox mailbox))
	  (when (imap-ok-p (imap-send-command-wait
			    (list "APPEND " (imap-encode-string mailbox) " "
				  article)))
	    (cdr (imap-mailbox-get 'appenduid mailbox))))
      (let (res)
	(when (setq res (imap-mailbox-status mailbox '(uidvalidity uidnext)))
	  (when (imap-ok-p (imap-send-command-wait
			    (list "APPEND " (imap-encode-string mailbox) " "
				  article)))
	    res))))))


;; Internal functions.

(defun imap-send-command (command &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (if (not (listp command)) (setq command (list command)))
    (let ((tag (setq imap-tag (1+ imap-tag)))
	  cmdstr cmd)
      (setq cmdstr (concat (number-to-string imap-tag) " "))
      (while (setq cmd (pop command))
	(cond ((stringp cmd)
	       (setq cmdstr (concat cmdstr cmd)))
	      (t
	       (error "Unknown command type"))))
      (setq cmdstr (concat cmdstr imap-client-eol))
      (and imap-log
	   (with-current-buffer (get-buffer-create imap-log)
	     (imap-disable-multibyte)
	     (buffer-disable-undo)
	     (goto-char (point-max))
	     (insert cmdstr)))
      (process-send-string imap-process cmdstr)
      tag)))

(defun imap-wait-for-tag (tag &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (while (< imap-reached-tag tag)
      (or (and (not (memq (process-status imap-process) '(open run)))
	       (sit-for 1))
	  (accept-process-output imap-process 1)))
    (or (assq tag imap-failed-tags)
	'OK)))

(defun imap-sentinel (process string)
  (delete-process process))

(defun imap-find-next-line ()
  "Return point at end of current line, taking into account
literals. Return nil if no complete line has arrived."
  (when (re-search-forward (concat imap-server-eol "\\|{\\([0-9]+\\)}"
				   imap-server-eol)
			   nil t)
    (if (match-string 1)
	(if (< (point-max) (+ (point) (string-to-number (match-string 1))))
	    nil
	  (goto-char (+ (point) (string-to-number (match-string 1))))
	  (imap-find-next-line))
      (point))))

(defun imap-arrival-filter (proc string)
  "IMAP process filter."
  (with-current-buffer (process-buffer proc)
    (goto-char (point-max))
    (insert string)
    (and imap-log
	 (with-current-buffer (get-buffer-create imap-log)
	   (imap-disable-multibyte)
	   (buffer-disable-undo)
	   (goto-char (point-max))
	   (insert string)))
    (goto-char (point-min))
    (let (end)
      (while (setq end (imap-find-next-line))
	(save-restriction
	  (narrow-to-region (point-min) end)
	  (delete-backward-char (length imap-server-eol))
	  (goto-char (point-min))
	  (unwind-protect
	      (cond ((eq imap-state 'initial)
		     (imap-parse-greeting))
		    ((or (eq imap-state 'auth)
			 (eq imap-state 'nonauth)
			 (eq imap-state 'selected))
		     (imap-parse-response))
		    (t
		     (message "Unknown state %s in arrival filter" 
			      imap-state)))
	    (delete-region (point-min) (point-max))))))))


;; Imap parser.

;;   literal         = "{" number "}" CRLF *CHAR8
;;                       ; Number represents the number of CHAR8s

(defsubst imap-parse-literal ()
  (when (looking-at "{\\([0-9]+\\)}\r\n")
    (let ((pos (match-end 0))
	  (len (string-to-number (match-string 1))))
      (if (< (point-max) (+ pos len))
	  nil
	(goto-char (+ pos len))
	(buffer-substring pos (+ pos len))))))

;;   string          = quoted / literal
;;
;;   quoted          = DQUOTE *QUOTED-CHAR DQUOTE
;;
;;   QUOTED-CHAR     = <any TEXT-CHAR except quoted-specials> /
;;                     "\" quoted-specials
;;
;;   quoted-specials = DQUOTE / "\"
;;
;;   TEXT-CHAR       = <any CHAR except CR and LF>

(defsubst imap-parse-string ()
  (cond ((eq (char-after) ?\")
	 (read (current-buffer)))
	((eq (char-after) ?{)
	 (imap-parse-literal))))
  
;;   nil             = "NIL"
;;
;;   nstring         = string / nil

(defsubst imap-parse-nstring ()
  (or (imap-parse-string)
      (when (looking-at "NIL")
	(goto-char (+ (point) 3))
	nil)))

;;   astring         = atom / string
;;
;;   atom            = 1*ATOM-CHAR
;;
;;   ATOM-CHAR       = <any CHAR except atom-specials>
;;
;;   atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
;;                     quoted-specials
;;
;;   list-wildcards  = "%" / "*"
;;
;;   quoted-specials = DQUOTE / "\"

(defsubst imap-parse-astring ()
  (or (imap-parse-string)
      (symbol-name (read (current-buffer)))))

;;   mailbox         = "INBOX" / astring
;;                       ; INBOX is case-insensitive.  All case variants of
;;                       ; INBOX (e.g. "iNbOx") MUST be interpreted as INBOX
;;                       ; not as an astring.  An astring which consists of
;;                       ; the case-insensitive sequence "I" "N" "B" "O" "X"
;;                       ; is considered to be INBOX and not an astring.
;;                       ;  Refer to section 5.1 for further
;;                       ; semantic details of mailbox names.

(defsubst imap-parse-mailbox ()
  (let ((mailbox (imap-parse-astring)))
    (and (symbolp mailbox)
	 (setq mailbox (symbol-name mailbox)))
    (if (string-equal "INBOX" (upcase mailbox))
	"INBOX"
      mailbox)))

;;   greeting        = "*" SP (resp-cond-auth / resp-cond-bye) CRLF
;;
;;   resp-cond-auth  = ("OK" / "PREAUTH") SP resp-text
;;                       ; Authentication condition
;;
;;   resp-cond-bye   = "BYE" SP resp-text

(defun imap-parse-greeting ()
  "Parse a IMAP greeting."
  (cond ((looking-at "\\* OK ")
	 (setq imap-state 'nonauth))
	((looking-at "\\* PREAUTH ")
	 (setq imap-state 'auth))
	((looking-at "\\* BYE ")
	 (setq imap-state 'closed))))

;;   response        = *(continue-req / response-data) response-done
;;
;;   continue-req    = "+" SP (resp-text / base64) CRLF
;;
;;   response-data   = "*" SP (resp-cond-state / resp-cond-bye /
;;                     mailbox-data / message-data / capability-data) CRLF
;;
;;   response-done   = response-tagged / response-fatal
;;
;;   response-fatal  = "*" SP resp-cond-bye CRLF
;;                       ; Server closes connection immediately
;;
;;   response-tagged = tag SP resp-cond-state CRLF
;;
;;   resp-cond-state = ("OK" / "NO" / "BAD") SP resp-text
;;                       ; Status condition
;;
;;   resp-cond-bye   = "BYE" SP resp-text
;;
;;   mailbox-data    =  "FLAGS" SP flag-list /
;;  		        "LIST" SP mailbox-list /
;;                      "LSUB" SP mailbox-list /
;;		        "SEARCH" *(SP nz-number) /
;;                      "STATUS" SP mailbox SP "("
;;	                      [status-att SP number *(SP status-att SP number)] ")" /
;;                      number SP "EXISTS" /
;;		        number SP "RECENT"
;;
;;   message-data    = nz-number SP ("EXPUNGE" / ("FETCH" SP msg-att))
;;
;;   capability-data = "CAPABILITY" *(SP capability) SP "IMAP4rev1"
;;                     *(SP capability)
;;                       ; IMAP4rev1 servers which offer RFC 1730
;;                       ; compatibility MUST list "IMAP4" as the first
;;                       ; capability.

(defun imap-parse-response ()
  "Parse a IMAP command response."
  (let (token)
    (case (setq token (read (current-buffer)))
      (+ (imap-parse-continue-req))
      (* (case (prog1 (setq token (read (current-buffer)))
		 (or (eobp) (forward-char)))
	   (OK         (imap-parse-resp-text))
	   (NO         (imap-parse-resp-text))
	   (BAD        (imap-parse-resp-text))
	   (BYE        (imap-parse-resp-text))
	   (FLAGS      (imap-mailbox-put 'flags (imap-parse-flag-list)))
	   (LIST       (imap-parse-data-list 'list))
	   (LSUB       (imap-parse-data-list 'lsub))
	   (SEARCH     (imap-mailbox-put 
			'search 
			(read (concat "(" (buffer-substring (point) (point-max)) ")"))))
	   (STATUS     (imap-parse-status))
	   (CAPABILITY (setq imap-capability 
			     (read (concat "(" (buffer-substring (point) (point-max))
					   ")"))))
	   (ACL        (imap-parse-acl))
	   (t       (case (prog1 (read (current-buffer))
			    (or (eobp) (forward-char)))
		      (EXISTS  (imap-mailbox-put 'exists token))
		      (RECENT  (imap-mailbox-put 'recent token))
		      (EXPUNGE t)
		      (FETCH   (imap-parse-fetch token))
		      (t       (message "Garbage: %s" (buffer-string)))))))
      (t (let (status)
	   (case (prog1 (setq status (read (current-buffer)))
		   (or (eobp) (forward-char)))
	   (OK  (progn
		  (setq imap-reached-tag (max imap-reached-tag token))
		  (imap-parse-resp-text)))
	   (NO  (progn
		  (setq imap-reached-tag (max imap-reached-tag token))
		  (save-excursion
		    (imap-parse-resp-text))
		  (let (code text)
		    (when (eq (char-after) ?\[)
		      (setq code (buffer-substring (point) (search-forward "]")))
		      (or (eobp) (forward-char)))
		    (setq text (buffer-substring (point) (point-max)))
		    (push (list token status code text) imap-failed-tags))))
	   (BAD (progn
		  (setq imap-reached-tag (max imap-reached-tag token))
		  (save-excursion
		    (imap-parse-resp-text))
		  (let (code text)
		    (when (eq (char-after) ?\[)
		      (setq code (buffer-substring (point) (search-forward "]")))
		      (or (eobp) (forward-char)))
		    (setq text (buffer-substring (point) (point-max)))
		    (push (list token status code text) imap-failed-tags)
		    (message "Internal error, tag %s status %s code %s text %s"
			     token status code text))))
	   (t   (message "Garbage after tag: %s" (buffer-string)))))))))

;;   resp-text       = ["[" resp-text-code "]" SP] text
;;
;;   text            = 1*TEXT-CHAR
;;
;;   TEXT-CHAR       = <any CHAR except CR and LF>

(defun imap-parse-resp-text ()
  (imap-parse-resp-text-code))

;;   resp-text-code  = "ALERT" /
;;                     "BADCHARSET [SP "(" astring *(SP astring) ")" ] /
;;                     "NEWNAME" SP string SP string / 
;;		       "PARSE" /
;;                     "PERMANENTFLAGS" SP "(" 
;;                               [flag-perm *(SP flag-perm)] ")" /
;;                     "READ-ONLY" / 
;;		       "READ-WRITE" / 
;;	 	       "TRYCREATE" /
;;                     "UIDNEXT" SP nz-number / 
;;		       "UIDVALIDITY" SP nz-number /
;;                     "UNSEEN" SP nz-number /
;;                     resp-text-atom [SP 1*<any TEXT-CHAR except "]">]
;;
;;   resp_code_apnd  = "APPENDUID" SPACE nz_number SPACE uniqueid
;;
;;   flag-perm       = flag / "\*"
;;
;;   flag            = "\Answered" / "\Flagged" / "\Deleted" /
;;                     "\Seen" / "\Draft" / flag-keyword / flag-extension
;;                       ; Does not include "\Recent"
;;
;;   flag-extension  = "\" atom
;;                       ; Future expansion.  Client implementations
;;                       ; MUST accept flag-extension flags.  Server
;;                       ; implementations MUST NOT generate
;;                       ; flag-extension flags except as defined by
;;                       ; future standard or standards-track
;;                       ; revisions of this specification.
;;
;;   flag-keyword    = atom
;;
;;   resp-text-atom  = 1*<any ATOM-CHAR except "]">

(defun imap-parse-resp-text-code ()
  (when (eq (char-after) ?\[)
    (or (eobp) (forward-char))
    (cond ((search-forward "PERMANENTFLAGS " nil t)
	   (imap-mailbox-put 'permanentflags (imap-parse-flag-list)))
	  ((search-forward "UIDNEXT " nil t)
	   (imap-mailbox-put 'uidnext (read (current-buffer))))
	  ((search-forward "UNSEEN " nil t)
	   (imap-mailbox-put 'unseen (read (current-buffer))))
	  ((looking-at "UIDVALIDITY \\([0-9]+\\)")
	   (imap-mailbox-put 'uidvalidity (match-string 1)))
	  ((search-forward "READ-ONLY" nil t)
	   (imap-mailbox-put 'read-only t))
	  ((search-forward "COPYUID" nil t)
	   t)
	  ((search-forward "NEWNAME " nil t)
	   (let (oldname newname)
	     (setq oldname (imap-parse-string))
	     (or (eobp) (forward-char))
	     (setq newname (imap-parse-string))
	     (imap-mailbox-put 'newname newname oldname)))
	  ((search-forward "TRYCREATE" nil t)
	   (imap-mailbox-put 'trycreate t imap-current-target-mailbox))
	  ((looking-at "APPENDUID \\([0-9]+\\) \\([0-9]+\\)")
	   (imap-mailbox-put 'appenduid 
			     (list (match-string 1)
				   (string-to-number (match-string 2)))
			     imap-current-target-mailbox))
	  ((search-forward "ALERT] " nil t)
	   (message "Imap server %s information: %s" imap-server
		    (buffer-substring (point) (point-max)))))))

;;   mailbox-list    = "(" [mbx-list-flags] ")" SP
;;                      (DQUOTE QUOTED-CHAR DQUOTE / nil) SP mailbox
;;
;;   mbx-list-flags  = *(mbx-list-oflag SP) mbx-list-sflag
;;                     *(SP mbx-list-oflag) /
;;                     mbx-list-oflag *(SP mbx-list-oflag)
;;
;;   mbx-list-oflag  = "\Noinferiors" / flag-extension
;;                       ; Other flags; multiple possible per LIST response
;;
;;   mbx-list-sflag  = "\Noselect" / "\Marked" / "\Unmarked"
;;                       ; Selectability flags; only one per LIST response
;;
;;   QUOTED-CHAR     = <any TEXT-CHAR except quoted-specials> /
;;                     "\" quoted-specials
;;
;;   quoted-specials = DQUOTE / "\"

(defun imap-parse-data-list (type)
  (let (flags delimiter mailbox)
    (setq flags (imap-parse-flag-list))
    (when (looking-at " NIL\\| \"\\(.\\)\"")
      (setq delimiter (match-string 1))
      (goto-char (1+ (match-end 0)))
      (when (setq mailbox (imap-parse-mailbox))
	(imap-mailbox-put type t mailbox)
	(imap-mailbox-put 'list-flags flags mailbox)
	(imap-mailbox-put 'delimiter delimiter mailbox)))))

;;   msg-att         = "(" (msg-att-dynamic / msg-att-static)
;;                      *(SP (msg-att-dynamic / msg-att-static)) ")"
;;
;;   msg-att-dynamic = "FLAGS" SP "(" [flag-fetch *(SP flag-fetch)] ")"
;;                       ; MAY change for a message
;;
;;   flag-fetch      = flag / "\Recent"
;;
;;   flag            = "\Answered" / "\Flagged" / "\Deleted" /
;;                     "\Seen" / "\Draft" / flag-keyword / flag-extension
;;                       ; Does not include "\Recent"
;;
;;   flag-keyword    = atom
;;
;;   flag-extension  = "\" atom
;;                       ; Future expansion.  Client implementations
;;                       ; MUST accept flag-extension flags.  Server
;;                       ; implementations MUST NOT generate
;;                       ; flag-extension flags except as defined by
;;                       ; future standard or standards-track
;;                       ; revisions of this specification.
;;
;;   msg-att-static  = "ENVELOPE" SP envelope / 
;;                     "INTERNALDATE" SP date-time /
;;                     "RFC822" [".HEADER" / ".TEXT"] SP nstring /
;;                     "RFC822.SIZE" SP number /
;;                     "BODY" ["STRUCTURE"] SP body /
;;                     "BODY" section ["<" number ">"] SP nstring /
;;                     "UID" SP uniqueid
;;                       ; MUST NOT change for a message
;;
;;   section         = "[" [section-spec] "]"
;;
;;   section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
;;                     "TEXT"
;;                       ; top-level or MESSAGE/RFC822 part
;;
;;   section-part    = nz-number *("." nz-number)
;;                       ; body part nesting
;;
;;   section-spec    = section-msgtext / (section-part ["." section-text])
;;
;;   section-text    = section-msgtext / "MIME"
;;                       ; text other than actual body part (headers, etc.)
;;
;;   date-time       = DQUOTE date-day-fixed "-" date-month "-" date-year
;;                     SP time SP zone DQUOTE
;;
;;   uniqueid        = nz-number
;;                       ; Strictly ascending

(defun imap-parse-fetch (response)
  (when (eq (char-after) ?\()
    (let (uid flags envelope internaldate rfc822 rfc822header rfc822text 
	      rfc822size body bodystructure)
      (while (not (eq (char-after) ?\)))
	(or (eobp) (forward-char))
	(let ((token (read (current-buffer))))
	  (or (eobp) (forward-char))
	  (cond ((eq token 'UID)
		 (setq uid (read (current-buffer))))
		((eq token 'FLAGS)
		 (setq flags (imap-parse-flag-list)))
		((eq token 'ENVELOPE)
		 (setq envelope (imap-parse-envelope)))
		((eq token 'INTERNALDATE)
		 (setq internaldate (imap-parse-string)))
		((eq token 'RFC822)
		 (setq rfc822 (imap-parse-nstring)))
		((eq token 'RFC822.HEADER)
		 (setq rfc822header (imap-parse-nstring)))
		((eq token 'RFC822.TEXT)
		 (setq rfc822text (imap-parse-nstring)))
		((eq token 'RFC822.SIZE)
		 (setq rfc822size (read (current-buffer))))
		((eq token 'BODY)
		 (setq body (imap-parse-body)))
		((eq token 'BODYSTRUCTURE)
		 (setq bodystructure (imap-parse-body))))))
      (if (not uid)
	  (message "Skipping UID-less fetch response (broken server?)...")
	(setq imap-current-message uid)
	(imap-message-put uid 'UID uid)
	(and flags (imap-message-put uid 'FLAGS flags))
	(and envelope (imap-message-put uid 'ENVELOPE envelope))
	(and internaldate (imap-message-put uid 'INTERNALDATE internaldate))
	(and rfc822 (imap-message-put uid 'RFC822 rfc822))
	(and rfc822header (imap-message-put uid 'RFC822.HEADER rfc822header))
	(and rfc822text (imap-message-put uid 'RFC822.TEXT rfc822text))
	(and rfc822size (imap-message-put uid 'RFC822.SIZE rfc822size))
	(and body (imap-message-put uid 'BODY body))
	(and bodystructure (imap-message-put uid 'BODYSTRUCTURE bodystructure))
	(run-hooks 'imap-fetch-data-hook)))))

;;   mailbox-data    =  ...
;;                      "STATUS" SP mailbox SP "("
;;	                      [status-att SP number 
;;                            *(SP status-att SP number)] ")"
;;                      ...
;;
;;   status-att      = "MESSAGES" / "RECENT" / "UIDNEXT" / "UIDVALIDITY" /
;;                     "UNSEEN"

(defun imap-parse-status ()
  (let ((mailbox (imap-parse-mailbox)))
    (when (and mailbox (search-forward "(" nil t))
      (while (not (eq (char-after) ?\)))
	(let ((token (read (current-buffer))))
	  (cond ((eq token 'MESSAGES)
		 (imap-mailbox-put 'messages (read (current-buffer)) mailbox))
		((eq token 'RECENT)
		 (imap-mailbox-put 'recent (read (current-buffer)) mailbox))
		((eq token 'UIDNEXT)
		 (imap-mailbox-put 'uidnext (read (current-buffer)) mailbox))
		((eq token 'UIDVALIDITY)
		 (and (looking-at " \\([0-9]+\\)")
		      (imap-mailbox-put 'uidvalidity (match-string 1) mailbox)
		      (goto-char (match-end 1))))
		((eq token 'UNSEEN)
		 (imap-mailbox-put 'unseen (read (current-buffer)) mailbox))
		(t
		 (message "Unknown status data %s in mailbox %s ignored" 
			  token mailbox))))))))

;;   acl_data        ::= "ACL" SPACE mailbox *(SPACE identifier SPACE
;;                        rights)
;;
;;   identifier      ::= astring
;;
;;   rights          ::= astring

(defun imap-parse-acl ()
  (let ((mailbox (imap-parse-mailbox)))
    (when (eq (char-after) ?\ )
      (let (acl)
	(while (eq (char-after) ?\ )
	  (let (identifier rights)
	    (setq identifier (imap-parse-astring))
	    (when (equal (char-after) ?\ )
	      (setq rights (imap-parse-astring))
	      (setq acl (append acl (list (cons identifier rights)))))))
	(imap-mailbox-put 'acl acl mailbox)))))

;;   flag-list       = "(" [flag *(SP flag)] ")"
;;
;;   flag            = "\Answered" / "\Flagged" / "\Deleted" /
;;                     "\Seen" / "\Draft" / flag-keyword / flag-extension
;;                       ; Does not include "\Recent"
;;
;;   flag-keyword    = atom
;;
;;   flag-extension  = "\" atom
;;                       ; Future expansion.  Client implementations
;;                       ; MUST accept flag-extension flags.  Server
;;                       ; implementations MUST NOT generate
;;                       ; flag-extension flags except as defined by
;;                       ; future standard or standards-track
;;                       ; revisions of this specification.

(defun imap-parse-flag-list ()
  (let ((str (buffer-substring (point) (search-forward ")" nil t)))
	pos)
    (while (setq pos (string-match "\\\\" str (and pos (+ 2 pos))))
      (setq str (replace-match "\\\\" nil t str)))
    (mapcar 'symbol-name (read str))))

;;   envelope        = "(" env-date SP env-subject SP env-from SP env-sender SP
;;                     env-reply-to SP env-to SP env-cc SP env-bcc SP
;;                     env-in-reply-to SP env-message-id ")"
;;
;;   env-bcc         = "(" 1*address ")" / nil
;;
;;   env-cc          = "(" 1*address ")" / nil
;;
;;   env-date        = nstring
;;
;;   env-from        = "(" 1*address ")" / nil
;;
;;   env-in-reply-to = nstring
;;
;;   env-message-id  = nstring
;;
;;   env-reply-to    = "(" 1*address ")" / nil
;;
;;   env-sender      = "(" 1*address ")" / nil
;;
;;   env-subject     = nstring
;;
;;   env-to          = "(" 1*address ")" / nil

(defun imap-parse-envelope ()
  ;; xxx: does not handle literals
  (read (current-buffer)))

;;   body            = "(" body-type-1part / body-type-mpart ")"
;;
;;   body-extension  = nstring / number /
;;                      "(" body-extension *(SP body-extension) ")"
;;                       ; Future expansion.  Client implementations
;;                       ; MUST accept body-extension fields.  Server
;;                       ; implementations MUST NOT generate
;;                       ; body-extension fields except as defined by
;;                       ; future standard or standards-track
;;                       ; revisions of this specification.
;;
;;   body-ext-1part  = body-fld-md5 [SP body-fld-dsp [SP body-fld-lang
;;                     *(SP body-extension)]]
;;                       ; MUST NOT be returned on non-extensible
;;                       ; "BODY" fetch
;;
;;   body-ext-mpart  = body-fld-param [SP body-fld-dsp [SP body-fld-lang
;;                     *(SP body-extension)]]
;;                       ; MUST NOT be returned on non-extensible
;;                       ; "BODY" fetch
;;
;;   body-fields     = body-fld-param SP body-fld-id SP body-fld-desc SP
;;                     body-fld-enc SP body-fld-octets
;;
;;   body-fld-desc   = nstring
;;
;;   body-fld-dsp    = "(" string SP body-fld-param ")" / nil
;;
;;   body-fld-enc    = (DQUOTE ("7BIT" / "8BIT" / "BINARY" / "BASE64"/
;;                     "QUOTED-PRINTABLE") DQUOTE) / string
;;
;;   body-fld-id     = nstring
;;
;;   body-fld-lang   = nstring / "(" string *(SP string) ")"
;;
;;   body-fld-lines  = number
;;
;;   body-fld-md5    = nstring
;;
;;   body-fld-octets = number
;;
;;   body-fld-param  = "(" string SP string *(SP string SP string) ")" / nil
;;
;;   body-type-1part = (body-type-basic / body-type-msg / body-type-text)
;;                     [SP body-ext-1part]
;;
;;   body-type-basic = media-basic SP body-fields
;;                       ; MESSAGE subtype MUST NOT be "RFC822"
;;
;;   body-type-mpart = 1*body SP media-subtype
;;                     [SP body-ext-mpart]
;;
;;   body-type-msg   = media-message SP body-fields SP envelope
;;                     SP body SP body-fld-lines
;;
;;   body-type-text  = media-text SP body-fields SP body-fld-lines
;;
;;   media-basic     = ((DQUOTE ("APPLICATION" / "AUDIO" / "IMAGE" / "MESSAGE" /
;;                     "VIDEO") DQUOTE) / string) SP media-subtype
;;                       ; Defined in [MIME-IMT]

;; ("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 7415 213)
;; (("TEXT" "PLAIN" ("CHARSET" "US-ASCII") NIL NIL "7BIT" 9 2)(("TEXT" "PLAIN" ("CHARSET" "US-ASCII") NIL NIL "7BIT" 8 2)("IMAGE" "GIF" NIL NIL NIL "BASE64" 2802) "ALTERNATIVE")("TEXT" "PLAIN" NIL NIL NIL "7BIT" 7 1) "MIXED")
;; (("IMAGE" "JPEG" NIL NIL NIL "BASE64" 1222)("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 2 1)(("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 21 2)("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 20 2)(("IMAGE" "JPEG" NIL NIL NIL "BASE64" 1222) "MIXED") "ALTERNATIVE")("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 111 5) "MIXED")

(defun imap-parse-body ()
  ;; xxx: does not handle literals
  (read (current-buffer)))

(when imap-debug ; (untrace-all)
  (require 'trace)
  (buffer-disable-undo (get-buffer-create imap-debug))
  (mapc (lambda (f) (trace-function-background f imap-debug)) 
        '(
imap-encode-string
imap-read-passwd
imap-kerberos4s-p
imap-kerberos4-open
imap-ssl-p
imap-ssl-open-1
imap-ssl-open
imap-network-p
imap-network-open
imap-interactive-login
imap-kerberos4a-p
imap-kerberos4-auth
imap-cram-md5-p
imap-cram-md5-auth
imap-login-p
imap-login-auth
imap-anonymous-p
imap-anonymous-open
imap-open-1
imap-open
imap-opened
imap-authenticate
imap-close
imap-capability
imap-namespace
imap-send-command-wait
imap-mailbox-put
imap-mailbox-get
imap-mailbox-map
imap-mailbox-select
imap-mailbox-unselect
imap-mailbox-close
imap-mailbox-create
imap-mailbox-lsub
imap-mailbox-list
imap-mailbox-subscribe
imap-mailbox-unsubscribe
imap-mailbox-status
imap-mailbox-acl-get
imap-mailbox-acl-set
imap-mailbox-acl-delete
imap-message-put
imap-message-get
imap-message-map
imap-search
imap-message-flags-set
imap-message-flags-del
imap-message-flags-add
imap-message-copy
imap-message-append
imap-send-command
imap-wait-for-tag
imap-sentinel
imap-find-next-line
imap-arrival-filter
imap-parse-greeting
imap-parse-response
imap-parse-resp-text
imap-parse-resp-text-code
imap-parse-data-list
imap-parse-fetch
imap-parse-status
imap-parse-acl
imap-parse-flag-list
imap-parse-envelope
imap-parse-body
	  )))
	
(provide 'imap)
